;;; snow.lisp
;;;
;;; Copyright (C) 2008-2009 Alessio Stalla
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU General Public License
;;; as published by the Free Software Foundation; either version 2
;;; of the License, or (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
;;;
;;; As a special exception, the copyright holders of this library give you
;;; permission to link this library with independent modules to produce an
;;; executable, regardless of the license terms of these independent
;;; modules, and to copy and distribute the resulting executable under
;;; terms of your choice, provided that you also meet, for each linked
;;; independent module, the terms and conditions of the license of that
;;; module.  An independent module is a module which is not derived from
;;; or based on this library.  If you modify this library, you may extend
;;; this exception to your version of the library, but you are not
;;; obligated to do so.  If you do not wish to do so, delete this
;;; exception statement from your version.

(in-package :snow)

(defvar *parent* nil)

(defun dashed->camelcased (string-designator)
  (let ((str (string string-designator))
	(last-was-dash-p nil))
    (with-output-to-string (out)
      (map nil (lambda (ch)
		 (if (char= #\- ch)
		     (setq last-was-dash-p t)
		     (if last-was-dash-p
			 (progn
			   (princ (char-upcase ch) out)
			   (setq last-was-dash-p nil))
			 (princ (char-downcase ch) out))))
	   str))))

(defgeneric widget-property (widget name)
  (:documentation "Retrieves the value of a widget's property. Widget properties names are dependent on the GUI backend and cannot be used portably across different GUI libraries."))
(defgeneric (setf widget-property) (value widget name)
  (:documentation "Sets the value of a widget's property. Widget properties names are dependent on the GUI backend and cannot be used portably across different GUI libraries."))

(defmethod (setf widget-property) (value widget name)
  (setf (jproperty-value widget (if (stringp name)
				    name
				    (dashed->camelcased name)))
	value))

(defmethod widget-property (widget name)
  (jproperty-value widget (if (stringp name)
			      name
			      (dashed->camelcased name))))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun map-keys (fn arglist &key (filter-if (constantly nil)))
  "Maps _fn_ over _arglist_, returning a new list containing the results of applying _fn_ on each key-value pair, in the same order. As an exception to this, when _fn_ returns a value V such that (funcall _filter-if_ V) is true, the value is not included in the returned list."
  (when arglist
    (let (key result)
      (loop
	 :for x :in arglist :for even := t :then (not even)
	 :if even :do (setq key x)
	 :else :do (let ((value (funcall fn key x)))
		     (unless (funcall filter-if value)
		       (push value result))))
      (nreverse result)))))

(defmacro set-widget-properties (widget &rest props)
  (with-unique-names (widget-var)
    `(let ((,widget-var ,widget))
       ,@(map-keys (lambda (key value)
		     `(setf (widget-property ,widget-var ,key) ,value))
		   props))))

(defgeneric bind-widget (widget binding)
  (:documentation "Connects a widget to a data binding. The framework automatically chooses which property of the widget to connect."))

(definterface make-layout-manager *gui-backend* (widget type &rest args))

(definterface (setf layout-manager) *gui-backend* (lm widget))

(defun setup-container-widget (self &key (layout-manager :default) &allow-other-keys)
  (setf (layout-manager self)
	(apply #'make-layout-manager self
	       (ensure-list (or layout-manager :default)))))

(defun generate-default-children-processing-code (id children)
  (let ((code
	 (loop
	    :for form :in children
	    :collect (if (listp form)
			 (cond 
			   ((get (car form) 'widget-p)
			    `(let ((*parent* self)) ,form))
			   (t `(let ((*parent* nil)) ,form)))
			 form))))
    (if id
	`((let ((,id self))
	    ,@code))
	code)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun common-widget-args ()
    '(layout binding (enabled-p t) (visible-p t) location size))
  (defun common-widget-args-declarations ()
    (let ((arg-names (mapcar (lambda (x) (if (atom x) x (car x)))
			     (common-widget-args))))
    `((declare (ignorable ,@arg-names)))))
  (defun filter-arglist (args filtered-keys)
    (loop
       :for key :in args :by #'cddr
       :for value :in (cdr args) by #'cddr
       :when (not (member key filtered-keys))
       :collect key :and
       :collect value))
  (defun filter-widget-args (args)
    "Eliminates widget arguments processed by common-widget-setup; else, they would be evaluated twice in the macro expansion."
    (filter-arglist args '(:id :layout :binding :enabled-p :visible-p :location
			   :layout-manager :size))))

(defun common-widget-setup (self layout binding enabled-p visible-p
			    location size)
  (setup-widget self :layout layout :binding binding :enabled-p enabled-p
		:visible-p visible-p :location location :size size))

(defun setup-widget (self &key layout binding (enabled-p t) (visible-p t)
		     location size &allow-other-keys)
  (when *parent* (add-child self *parent* layout))
  (setf (widget-enabled-p self) enabled-p)
  (setf (widget-visible-p self) visible-p)
  (when location (setf (widget-location self) location))
  (when binding (bind-widget self binding))
  (when size (setf (widget-size self) size)))

#+emacs (put 'define-widget-macro 'lisp-indent-function 3)
#+emacs (put 'define-widget 'lisp-indent-function 3)
#+emacs (put 'define-container-widget 'lisp-indent-function 3)

(defmacro define-widget-macro (name arglist constructor &body body)
  `(progn
     (defmacro ,name (,@arglist)
       `(let ((self ,,constructor)) ;The lexical variable self is always bound to the current widget.
	  ,,@body
	  self))
     (setf (get ',name 'widget-p) t)))

;;Experimental - not working right now
(defmacro define-widget-function (name arglist constructor &body body)
  `(progn
     (defun ,name (,@arglist)
       `(let ((self ,,constructor)) ;The lexical variable self is always bound to the current widget.
	  ,,@body
	  self))
     (setf (get ',name 'widget-p) t)))

(define-widget-macro with-widget
    ((widget &rest args &key id layout binding (enabled-p t) (visible-p t)
	     location size)
     &body body)
    `(dont-add ,widget)
  `(progn
     ,@(generate-default-children-processing-code id body)
     (common-widget-setup self ,layout ,binding ,enabled-p ,visible-p ,location ,size)))

(define-widget-macro child
    (widget &rest args &key layout binding (enabled-p t) (visible-p t)
	    location size)
    widget
  `(setup-widget ,@args))

(defmacro define-widget (name keys constructor &body body)
  "Convenience macro for defining a widget."
  (with-unique-names (args)
    `(define-widget-macro ,name
	 (&rest ,args &key ,@(common-widget-args) ,@keys)
	 `(funcall (lambda (&rest args) ;;to evaluate args only once
		     (let ((self (apply (function ,',constructor) args)))
		       (apply #'setup-widget self args)
		       self))
		   ,@,args)
       `(progn
	  ,,@body))))

(defmacro define-container-widget (name keys constructor &body body)
  "Convenience macro for defining a container widget."
  (with-unique-names (args macro-body)
    `(define-widget-macro ,name
	 ((&rest ,args &key id ,@(common-widget-args) layout-manager ,@keys)
	  &body ,macro-body)
	 `(funcall (lambda (&rest args) ;;to evaluate args only once
		     (let ((self (apply (function ,',constructor) args)))
		       (apply #'setup-widget self args)
		       (apply #'setup-container-widget self args)
		       self))
		   ;;remove id because it must not be evaluated
		   ,@(filter-arglist ,args '(:id)))
       `(progn
	  ,(progn ,@body) ;Bug in ABCL? ,,@body fails when body is NIL: Wrong number of arguments for CONS - it generates (cons (append (generate...) (apply...)))
	  ,@(generate-default-children-processing-code id ,macro-body)))))

(defmacro auto-add-children (&body body)
  `(let ((*parent* self))
     ,@body))

(defmacro dont-add (&body body)
  `(let ((*parent* nil))
     ,@body))

(definterface call-in-gui-thread *gui-backend* (fn)
  "Arranges <fn> to be called from a thread in which it is safe to create GUI components (for example, the Event Dispatching Thread in Swing).")

(defvar *dynamic-environment* nil)

(defmacro with-gui ((&optional (gui-backend '*gui-backend*)) &body body)
  (with-unique-names (gui-backend-var package-var debugger-hook-var
		      dynamic-environment terminal-io-var)
    `(let* ((,gui-backend-var ,gui-backend)
	    (*gui-backend* ,gui-backend-var)
	    (,package-var *package*)
	    (,debugger-hook-var *debugger-hook*)
	    (,terminal-io-var *terminal-io*)) ;;Etc...
       (dynamic-wind
	(let ((*gui-backend* ,gui-backend-var)
	      (*package* ,package-var)
	      (*debugger-hook* ,debugger-hook-var)
	      (*terminal-io* ,terminal-io-var))
	  (proceed
	   (let ((,dynamic-environment (capture-dynamic-environment)))
	     (call-in-gui-thread
	      (lambda ()
		(with-dynamic-environment (,dynamic-environment)
		  (let ((*dynamic-environment* ,dynamic-environment))
		    ,@body)))))))))))

;;Common Interfaces
(defvar *gui-backend* :swing "Variable used to determine the GUI backend, and thus interface implementation, to use. Defaults to :swing.")

(definterface add-child *gui-backend* (child &optional (parent *parent*) layout-constraints))

(definterface widget-enabled-p *gui-backend* (widget))

(definterface (setf widget-enabled-p) *gui-backend* (value widget))

(definterface widget-visible-p *gui-backend* (widget))

(definterface (setf widget-visible-p) *gui-backend* (value widget))

(definterface (setf widget-location) *gui-backend* (value widget))

(definterface (setf widget-size) *gui-backend* (value widget))

(definterface dispose *gui-backend* (obj))

(definterface show *gui-backend* (obj))

(definterface hide *gui-backend* (obj))

(definterface pack *gui-backend* (window))

;;Windows
(definterface make-frame *gui-backend* (&key menu-bar title on-close
					&allow-other-keys))

(define-container-widget frame (menu-bar title on-close) make-frame)

(definterface make-dialog *gui-backend*
  (&key parent title modal-p visible-p &allow-other-keys))

(define-widget-macro dialog
    ((&rest args &key id layout binding (enabled-p t) (visible-p t) location
	    size layout-manager parent title modal-p visible-p)
     &body body)
    `(funcall (lambda (&rest args) ;;to evaluate args only once
		(let ((self (apply (function make-dialog) args)))
		  (apply #'setup-widget self `(:visible-p nil ,@args))
		  (apply #'setup-container-widget self args)
		  self))
	      ;;remove id because it must not be evaluated
	      ;;and visible-p because it must be set last
	      ,@(filter-arglist args '(:id :visible-p)))
  `(progn
     ,@(generate-default-children-processing-code id body)
     (setf (widget-visible-p self) ,visible-p)))

#|
(define-container-widget dialog (parent title modal-p)
  make-dialog)|#

;;Menus
(definterface make-menu-bar *gui-backend* (&key &allow-other-keys))

(define-container-widget menu-bar () make-menu-bar)

(definterface make-menu *gui-backend* (&key text &allow-other-keys))

(define-container-widget menu (text) make-menu)

(definterface make-menu-item *gui-backend*
  (&key text on-action &allow-other-keys))

(define-widget menu-item (text on-action) make-menu-item)

;;Panels
(definterface make-panel *gui-backend* (&key &allow-other-keys))

(define-container-widget panel () make-panel)

(defvar *tabs*)

(definterface make-tabs *gui-backend* (&key (wrap t) (tab-placement :top)
				       &allow-other-keys))

(define-widget-macro tabs
    ((&rest args
      &key id layout binding (enabled-p t) (visible-p t) location size (wrap t)
	   (tab-placement :top))
     &body body)
    `(make-tabs :wrap ,wrap :tab-placement ,tab-placement)
  `(let ((*tabs* self))
     (dont-add
       ,@(if id
	     `((let ((,id self))
		 ,@body))
	     body))
     (common-widget-setup self ,layout ,binding ,enabled-p ,visible-p
			  ,location ,size)))

(defmacro tab (name &body body)
  `(if *tabs*
       (add-child (progn ,@body) *tabs* ,name)
       (error "tab outside tabset: ~A" ,name)))

(definterface make-scroll-panel *gui-backend* (view))

(definterface scroll-panel-view *gui-backend* (self))

(definterface (setf scroll-panel-view) *gui-backend* (view self))

(define-widget-macro scroll
    ((&rest args &key layout binding (enabled-p t) (visible-p t) location size) body)
    `(make-scroll-panel (dont-add ,body))
  `(setup-widget self ,@args))

(definterface make-split-panel *gui-backend*
  (child1 child2 &key (orientation :horizontal) smoothp))

(define-widget-macro split
    ((&rest args &key layout binding (enabled-p t) (visible-p t) location size orientation smoothp)
     child1 child2)
    `(make-split-panel (dont-add ,child1) (dont-add ,child2)
		       :orientation ,orientation :smoothp ,smoothp)
  `(common-widget-setup self ,layout ,binding ,enabled-p ,visible-p ,location ,size))

(defmacro defwidget (name &rest args)
  (let* ((maker-sym (intern (concatenate 'string "MAKE-" (symbol-name name)))))
    `(progn
       (definterface ,maker-sym *gui-backend* (&key ,@args &allow-other-keys))
       (define-widget ,name (,@args &allow-other-keys) ,maker-sym))))

;;Buttons and similar

(defwidget button text on-action)

(defwidget check-box text selected-p)

;;Misc

(defwidget progress-bar value orientation (paint-border t) progress-string)

;;Text

(defwidget label text)

(defwidget text-field text)

(defwidget text-area text)

;;Lists

(defwidget list-widget model selected-index)

;;Trees

(defwidget tree model)